@@ -3,14 +3,16 @@
 
 ((window) => {
   const {
-    StringPrototypeReplace,
-    TypeError,
-    Promise,
     decodeURIComponent,
     Error,
+    ObjectPrototypeIsPrototypeOf,
+    Promise,
+    SafeArrayIterator,
+    StringPrototypeReplace,
+    TypeError,
   } = window.__bootstrap.primordials;
   const { build } = window.__bootstrap.build;
-  const { URL } = window.__bootstrap.url;
+  const { URLPrototype } = window.__bootstrap.url;
   let logDebug = false;
   let logSource = "JS";
 
@@ -25,7 +27,10 @@
     if (logDebug) {
       // if we destructure `console` off `globalThis` too early, we don't bind to
       // the right console, therefore we don't log anything out.
-      globalThis.console.log(`DEBUG ${logSource} -`, ...args);
+      globalThis.console.log(
+        `DEBUG ${logSource} -`,
+        ...new SafeArrayIterator(args),
+      );
     }
   }
 
@@ -93,7 +98,7 @@
   }
 
   function pathFromURL(pathOrUrl) {
-    if (pathOrUrl instanceof URL) {
+    if (ObjectPrototypeIsPrototypeOf(URLPrototype, pathOrUrl)) {
       if (pathOrUrl.protocol != "file:") {
         throw new TypeError("Must be a file URL.");
       }
